"""
    Python Zendesk is wrapper for the Zendesk API. This library provides an
    easy and flexible way for developers to communicate with their Zendesk
    account in their application.

    Notes:
    API THROTTLE is not handled in this library:
        From Zendesk: The throttle will be applied once an API consumer
        reaches a certain threshold in terms of a maximum of requests per
        minute. Most clients will never hit this threshold, but those that do,
        will get met by a HTTP 503 response code and a text body of
        "Number of allowed API requests per minute exceeded"

    TICKETS AS AN END-USER is not handled in this library:
        There are a number of API calls for working with helpdesk tickets as
        if you were the person (end-user) who originally created them. At the
        moment, this library is primarily for admins and agents to interact

    FORUMS, ENTRIES, POSTS have not yet been implemented

"""

__author__ = "Max Gutman <max@eventbrite.com>"
__version__ = "1.0.0"

import re
import urllib
import urllib2
import httplib2
import dict2xml

from endpoints import mapping_table


class ZendeskError(Exception):
    def __init__(self, msg, error_code=None):
        self.msg = msg
        self.error_code = error_code
        # Zendesk will throw a 401 response for un-authneticated call
        if self.error_code == 401:
            raise AuthenticationError(self.msg)

    def __str__(self):
        return repr('%s: %s' % (self.error_code, self.msg))


class AuthenticationError(ZendeskError):
	def __init__(self, msg):
		self.msg = msg

	def __str__(self):
		return repr(self.msg)


class XMLParseError(Exception):
	def __init__(self, msg):
		self.msg = msg

	def __str__(self):
		return repr(self.msg)


class Zendesk(object):
    """ Python API Wrapper for Zendesk"""

    def __init__(self, zendesk_url, zendesk_username=None,
                       zendesk_password=None, headers=None, client_args={}):
        """
        Instantiates an instance of Zendesk. Takes optional parameters for
        HTTP Basic Authentication

        Parameters:
        zendesk_url - https://company.zendesk.com (use http if not SSL enabled)
        zendesk_username - Specific to your Zendesk account (typically email)
        zendesk_password - Specific to your Zendesk account
        headers - Pass headers in dict form. This will override default.
        client_args - Pass arguments to http client in dict form.
            {'cache': False, 'timeout': 2}
        """
        self.xml_data = None

        # Set attributes necessary for API
        self.zendesk_url = zendesk_url
        self.zendesk_username = zendesk_username
        self.zendesk_password = zendesk_password

        # Set headers
        self.headers = headers
        if self.headers is None:
            self.headers = {
                'User-agent': 'Zendesk Python Library v%s' % __version__,
                'Content-Type': 'application/xml'
            }

        # Set http client and authentication
        self.client = httplib2.Http(**client_args)
        if self.zendesk_username is not None and self.zendesk_password is not None:
            self.client.add_credentials(
                self.zendesk_username,
                self.zendesk_password
            )

    def __getattr__(self, api_call):
        """
            Instead of writing out each API endpoint as a method here or
            binding the API endpoints at instance runttime, we can simply
            use an elegant Python technique to construct method execution on-
            demand. We simply provide a mapping table between Zendesk API calls
            and function names (with necessary parameters to replace
            embedded keywords on GET or xml data on POST/PUT requests).

            __getattr__() is used as callback method implemented so that
            when an object tries to call a method which is not defined here,
            it looks to find a relationship in the the mapping table.  The
            table provides the structure of the API call and parameters passed
            in the method will populate missing data.
            
            TODO:
                Should probably url-encode GET query parameters on replacement
        """
        def call(self, **kwargs):
            """ """
            api_map = mapping_table[api_call]
            method = api_map['method']
            path = api_map['path']
            status = api_map['status']
            # Body can be passed from xml_data or in args
            body = kwargs.get('xml_data') or self.xml_data
            # Substitute mustache placeholders with data from keywords
            url = re.sub(
                '\{\{(?P<m>[a-zA-Z_]+)\}\}',
                # Optional pagination parameters will default to blank
                lambda m: "%s" % kwargs.get(m.group(1), ''),
                self.zendesk_url + path
            )
            # Make an http request (data replacements are finalized)
            response, content = \
                self.client.request(
                    url,
                    method,
                    body=body,
                    headers=self.headers
                )
            # Use a response handler to determine success/fail
            return self._response_handler(response, content, status)

        # Missing method is also not defined in our mapping table
        if api_call not in mapping_table:
            raise AttributeError('Method "%s" Does Not Exist' % api_call)

        # Execute dynamic method and pass in keyword args as data to API call
        return call.__get__(self)

    @staticmethod
    def _response_handler(response, content, status):
        """ Handle response as callback

            If the response status is different from status defined in the
            mapping table, then we assume an error and raise proper exception

            Zendesk's response is sometimes the url of a newly created user/
            ticket/group/etc and they pass this through 'location'.  Otherwise,
            the body of 'content' has our response.
        """
        # Just in case
        if not response:
            raise ZendeskError('Response Not Found')
        response_status = int(response.get('status', 0))
        if response_status != status:
            raise ZendeskError(content, response_status)
        return response.get('location') or content or 'Success'

    @staticmethod
    def dict2xml(data_dict, pretty_print=True):
        """ dict2xml converts a Python dictionary into an XML string
            See examples for more complex formatting.
        """
        try:
            return dict2xml.tostring(data_dict, pretty_print)
        except Exception, error:
            # Let's bubble everything up as an XML parsing error
            raise XMLParseError(
                'Dictionary -> XML Conversion Failed: "%s"' % error
            )




"""
API MAPPING

"""

mapping_table = {
    # Rest API: Organizations
    'list_organizations': {
        'path': '/organizations.xml',
        'method': 'GET',
        'status': 200,
    },
    'show_organization': {
        'path': '/organizations/{{organization_id}}.xml',
        'method': 'GET',
        'status': 200,
    },
    'create_organization': {
        'path': '/organizations.xml',
        'method': 'POST',
        'status': 201,
    },
    'update_organization': {
        'path': '/organizations/{{organization_id}}.xml',
        'method': 'PUT',
        'status': 200,
    },
    'delete_organization': {
        'path': '/organizations/{{organization_id}}.xml',
        'method': 'DELETE',
        'status': 200,
    },
    # Rest API: Groups
    'list_groups': {
        'path': '/groups.xml',
        'method': 'GET',
        'status': 200,
    },
    'show_group': {
        'path': '/groups/{{group_id}}.xml',
        'method': 'GET',
        'status': 200,
    },
    'create_group': {
        'path': '/groups.xml',
        'method': 'POST',
        'status': 201,
    },
    'update_group': {
        'path': '/groups/{{group_id}}.xml',
        'method': 'PUT',
        'status': 200,
    },
    'delete_group': {
        'path': '/groups/{{group_id}}.xml',
        'method': 'DELETE',
        'status': 200,
    },
    # Rest API: Tickets
    'list_tickets': {
        'path': '/rules/{{view_id}}.xml?page={{page}}',
        'method': 'GET',
        'status': 200,
    },
    'show_ticket': {
        'path': '/tickets/{{ticket_id}}.xml',
        'method': 'GET',
        'status': 200,
    },
    'create_ticket': {
        'path': '/tickets.xml',
        'method': 'POST',
        'status': 201,
    },
    'update_ticket': {
        'path': '/tickets/{{ticket_id}}.xml',
        'method': 'PUT',
        'status': 200,
    },
    'comment_ticket': {
        'path': '/tickets/{{ticket_id}}.xml',
        'method': 'PUT',
        'status': 200,
    },
    'delete_ticket': {
        'path': '/tickets/{{ticket_id}}.xml',
        'method': 'DELETE',
        'status': 200,
    },
    # Rest API: Attachment
    'create_attachment': {
        'path': '/uploads.xml?filename={{filename}}&amp;token={{token}}',
        'method': 'POST',
        'status': 201,
    },
    # Rest API: Users
    'list_users': {
        'path': '/users.xml?page={{page}}',
        'method': 'GET',
        'status': 200,
    },
    'search_users': {
        'path': '/users.xml?query={{query}}&amp;role={{role}}&amp;page={{page}}',
        'method': 'GET',
    },
    'show_user': {
        'path': '/users/{{user_id}}.xml',
        'method': 'GET',
        'status': 200,
    },
    'create_user': {
        'path': '/users.xml',
        'method': 'POST',
        'status': 201,
    },
    'update_user': {
        'path': '/users/{{user_id}}.xml',
        'method': 'PUT',
        'status': 200,
    },
    'delete_user': {
        'path': '/users/{{user_id}}.xml',
        'method': 'DELETE',
        'status': 200,
    },
    'list_user_identities': {
        'path': '/users/{{user_id}}/user_identities.xml',
        'method': 'GET',
        'status': 200,
    },
    'add_user_email': {
        'path': '/users/{{user_id}}/user_identities.xml',
        'method': 'POST',
        'status': 201,
    },
    'add_twitter_handle': {
        'path': '/users/{{user_id}}/user_identities.xml',
        'method': 'POST',
        'status': 201,
    },
    'make_identity_primary': {
        'path': '/users/{{user_id}}/user_identities/{{identity_id}}/make_primary',
        'method': 'POST',
        'status': 200,
    },
    'delete_identity': {
        'path': '/users/{{user_id}}/user_identities/{{identity_id}}',
        'method': 'DELETE',
        'status': 200,
    },
    # Rest API: Tags
    'list_tags': {
        'path': '/tags.xml',
        'method': 'GET',
        'status': 200,
    },
    'list_assets': {
        'path': '/tags/{{tag_id}}.xml?for={{asset_type}}&amp;page={{page}}',
        'method': 'GET',
        'status': 200,
    },
    # Rest API: Ticket Fields
    'list_ticket_fields': {
        'path': '/ticket_fields.xml',
        'method': 'GET',
        'status': 200,
    },
    # Rest API: Macros
    'list_macros': {
        'path': '/macros.xml',
        'method': 'GET',
        'status': 200,
    },
    'evaluate_macro': {
        'path': '/macros/{{macro_id}}/apply.xml?ticket_id={{ticket_id}}',
        'method': 'POST',
        'status': 201,
    },
    # Rest API: Search
    'search': {
        'path': '/search.xml?query={{query}}&amp;page={{page}}',
        'method': 'GET',
        'status': 200,
    },
}


#!/usr/bin/env python2.5
# -*- coding: utf-8 -*-
# 
# Copyright 2010 Vojtech Rylko <vojta.rylko @ seznam.cz>
#

"""
Module for converting dict to xml.
"""
# inspired by http://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html.

__author__ = "xrylko00@stud.fit.vutbr.cz"

INDENTATION = 4     # for one level
ATTRIBUTE_PREFIX = "@"
TEXT = "#text"


def _test(dictionary):
    if not isinstance(dictionary, dict):
        raise TypeError("Type '%s' cannot be serialized." % \
                        type(dictionary).__name__)


def tostring(dictionary, pretty_print=False):
    """Convert dictionary to xml.
    
    Prefix `@` for attributes, `#text` for text.
    
    For examples see test/__main__.py

    @return: string
    """
    _test(dictionary)
    if dictionary == {}: return ''
    return "".join(_build_element(e, v, 0, pretty_print) for e, v in dictionary.items())
    
def to_etree(dictionary, etree):
    """Convert dictionary to etree structure.

    Prefix `@` for attributes, `#text` for text.
    
    For examples see test/__main__.py

    @return: string
    """
    _test(dictionary)
    if len(dictionary) > 1:
        raise TypeError("Dictionary can have only one root element.")
    if dictionary == {}: return etree.Element(None)
    r, v = dictionary.items()[0]
    return _build_element(r, v, etree=etree)
    
def _build_element(root, value, level=0, pretty_print=False, etree=None):
    """
    Make element with name `root` and with value `value`. It can
    contains attributes.
    
    Specials: starts with @ for attributes, #text for text.
    
    @return: string
    """
    if pretty_print: 
        newline = '\n'
        tabular = " "*(INDENTATION*level)
    else: 
        newline = ''
        tabular = ""

    if not value:
        if not etree : result = "<%s />%s" % (root, newline)
        else : result = etree.Element(str(root))
    elif isinstance(value, list):
        tabular = ""
        if not etree:
            return "".join([_build_element(root, r, level, pretty_print) for r in value])
        else:
            #raise ValueError("{root: [list,]} is not supported schema with etree")
            result = []
            for subelement in value:
                result.append(_build_element(root, subelement, etree=etree))
    elif isinstance(value, dict):
        values = []
        attribs = []
        for r, v in value.iteritems():
            if (isinstance(r, basestring) and r.startswith(ATTRIBUTE_PREFIX)):
                # its attribute
                attribs.append(_build_attrib(r,v, etree))
            else:
                # its content = element
                values.append(_build_element(r, v, level+1, pretty_print, etree))
        attr = ""
        if not etree:
            if attribs: 
                # prepare `attr` for attributes
                attr = " "+" ".join(attribs)
                
            if values:
                # result with content
                result = "<%s%s>%s%s</%s>%s" % \
                        (root, attr, newline, 
                         "".join(values), root, newline)
            else:
                # result without any content
                result = "<%s%s/>" % (root, attr)
        else: # etree
            result = etree.Element(str(root))
            for atr, val in attribs:
                result.attrib[atr] = val
            for value in values:
                if isinstance(value, basestring):
                    result.text = value
                elif isinstance(value, list):
                    for subval in value:
                        result.append(subval)
                else: result.append(value)

    elif root == TEXT:
        if value == '<\'">':
            result = quote_xml(value)
        else:
            result = quote_xml(value)
    else: 
        if not etree:
            result = "<%s>%s</%s>%s" % (root, quote_xml(value), root, newline)
        else:
            result = etree.Element(str(root))
            result.text = unicode(value)
    if etree : return result
    else : return  tabular + result


#
# Help Functions
#
def _build_attrib(name, value, etree=None):
    """
    Build attribute as 'name="value"'
    """
    if not etree: 
        return '%s=%s' % (name[1:], quote_attrib(value)) # remove first char @
    else:
        return (name[1:], unicode(value))

def quote_xml(inStr):
    """Author: dkuhlman@rexx.com, MIT Licence"""
    if not inStr:
        return ''
    s1 = (isinstance(inStr, basestring) and inStr or
          '%s' % inStr)
    s1 = s1.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    return s1

def quote_attrib(inStr):
    """Author: dkuhlman@rexx.com, MIT Licence"""
    s1 = (isinstance(inStr, basestring) and inStr or
          '%s' % inStr)
    s1 = s1.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    if '"' in s1:
        if "'" in s1:
            s1 = '"%s"' % s1.replace('"', "&quot;")
        else:
            s1 = "'%s'" % s1
    else:
        s1 = '"%s"' % s1
    return s1



import re
from zendesk import Zendesk

def get_id_from_url(url):
    match = re.match(r".*/(?P<identifier>\d+)\.xml", url)
    if match and match.group('identifier'):
        return match.group('identifier')


################################################################
## NEW CONNECTION CLIENT
################################################################
zendesk = Zendesk('https://yourcompany.zendesk.com', 'you@yourcompany.com', 'passwd')

################################################################
## TICKETS
################################################################

# List
zendesk.list_tickets(view_id=1) # Must have a view defined

# Create
new_ticket = {
    'ticket': {
        'requester-name': 'Howard Schultz',
        'requester-email': 'howard@starbucks.com',
        'subject':'My Starbucks coffee is cold!',
        'description': 'please reheat my coffee',
        'set-tags': 'coffee drinks',
        'ticket-field-entries': {
            '@type': 'array',
            'ticket-field-entry': [
                {'ticket-field-id': 1, 'value': 'venti'},
                {'ticket-field-id': 2, 'value': '$10'}
            ]
        },
    }
}
post_data = Zendesk.dict2xml(new_ticket)
ticket_url = zendesk.create_ticket(xml_data=post_data)
ticket_id = get_id_from_url(ticket_url)

# Show
zendesk.show_ticket(ticket_id=ticket_id)

# Delete
zendesk.delete_ticket(ticket_id=ticket_id)


################################################################
## ORGANIZATIONS
################################################################

# List
zendesk.list_organizations()

# Create
new_org = {
    'organization': {
        'name': 'Starbucks Corp'
    }
}
post_data = Zendesk.dict2xml(new_org)
org_url = zendesk.create_organization(xml_data=post_data)
org_id = get_id_from_url(org_url)

# Show
zendesk.show_organization(organization_id=org_id)

# Delete
zendesk.delete_organization(organization_id=org_id)


################################################################
## USERS (AGENTS)
################################################################

# List
zendesk.list_users()

# Create
new_user = {
    'user': {
        'name': 'Howard Schultz',
        'email': 'howard@starbucks.com',
        'roles': 4,
    }
}
post_data = Zendesk.dict2xml(new_user)
user_url = zendesk.create_user(xml_data=post_data)
user_id = get_id_from_url(user_url)

# Show
zendesk.show_user(group_id=user_id)

# Delete
zendesk.delete_user(group_id=user_id)


################################################################
## GROUPS
################################################################

# List
zendesk.list_groups()

# Create
new_group = {
    'group': {
        'name': 'Starbucks Group',
        'agents': {
            '@type': 'array',
            'agent': 123,
        }
    }
}
post_data = Zendesk.dict2xml(new_group)
group_url = zendesk.create_group(xml_data=post_data)
group_id = get_id_from_url(group_url)

# Show
zendesk.show_group(group_id=group_id)

# Delete
zendesk.delete_group(group_id=group_id)


################################################################
## TAGS
################################################################

# List
zendesk.list_tags()

# Show
zendesk.list_assets(tag_id=123, asset_type='event') # event | entry


################################################################
## TICKET TYPES
################################################################
zendesk.list_ticket_fields()


################################################################
## SEARCH
################################################################

# http://www.zendesk.com/api/search
# make sure to url-encode the query
results = zendesk.search(query='ticket+sort:desc', page=1)


#!/usr/bin/python

from distutils.core import setup

setup(
	# Basic package information.
	name = 'Zendesk',
	author = 'Max Gutman',
	version = '1.0.0',
	author_email = 'max@eventbrite.com',
	packages = ['zendesk'],
	include_package_data = True,
	install_requires = ['httplib2'],
	license='LICENSE.txt',
	url = 'http://github.com/eventbrite/zendesk/tree/master',
	keywords = 'zendesk api helpdesk',
	description = 'Python API Wrapper for Zendesk',
	classifiers = [
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: MIT License',
		'Topic :: Software Development :: Libraries :: Python Modules',
		'Topic :: Internet'
	],
)


